@@ -4,7 +4,7 @@ import numpy as np
 import os
 from numba import njit, prange
 
-from .maxcut_tfim_util import get_cut, init_theta, init_thresholds, low_width, maxcut_hamming_cdf, opencl_context, probability_by_hamming_weight
+from .maxcut_tfim_util import fix_cdf, get_cut, init_thresholds, maxcut_hamming_cdf, opencl_context, probability_by_hamming_weight
 
 IS_OPENCL_AVAILABLE = True
 try:
@@ -25,46 +25,43 @@ def local_repulsion_choice(G_m, max_weight, weights, n, m):
     """
 
     weights = weights.copy()
-    chosen = np.zeros(m, dtype=np.int32)   # store chosen indices
-    available = np.ones(n, dtype=np.bool_) # True = available, False = not
-    mask = np.zeros(n, dtype=np.bool_)
-    chosen_count = 0
+    used = np.zeros(n, dtype=np.bool_) # False = available, True = used
 
     for _ in range(m):
         # Count available
         total_w = 0.0
         for i in range(n):
-            if available[i]:
-                total_w += weights[i]
-        if total_w <= 0:
-            break
+            if used[i]:
+                continue
+            total_w += weights[i]
 
         # Normalize & sample
         r = np.random.rand()
         cum = 0.0
         node = -1
         for i in range(n):
-            if available[i]:
-                cum += weights[i] / total_w
-                if r < cum:
-                    node = i
-                    break
+            if used[i]:
+                continue
+            cum += weights[i]
+            if (total_w * r) < cum:
+                node = i
+                break
 
         if node == -1:
-            continue
+            node = 0
+            while used[node]:
+                node += 1
 
         # Select node
-        chosen[chosen_count] = node
-        chosen_count += 1
-        available[node] = False
-        mask[node] = True
+        used[node] = True
 
         # Repulsion: penalize neighbors
         for nbr in range(n):
-            if available[nbr]:
-                weights[nbr] *= 0.5 ** (G_m[node, nbr] / max_weight)  # tunable penalty factor
+            if used[nbr]:
+                continue
+            weights[nbr] *= max(2e-7, 1 - G_m[node, nbr] / max_weight)
 
-    return mask
+    return used
 
 
 @njit
@@ -81,7 +78,7 @@ def compute_energy(sample, G_m, n_qubits):
 def sample_for_solution(G_m, shots, thresholds, J_eff):
     n = len(G_m)
     max_weight = G_m.max()
-    weights = 1.0 / (1.0 + (2 ** -52) - J_eff)
+    weights = (1.0 / (1.0 + (2e-52) - J_eff)).astype(np.float64)
 
     solutions = np.empty((shots, n), dtype=np.bool_)
     energies = np.empty(shots, dtype=np.float32)
@@ -129,6 +126,78 @@ def init_J_and_z(G_m):
     return J_eff, degrees
 
 
+def run_sampling_opencl(G_m_np, thresholds_np, shots, n):
+    ctx = opencl_context.ctx
+    queue = opencl_context.queue
+    kernel = opencl_context.sample_for_solution_best_bitset_kernel
+
+    max_local_size = 64  # tune
+    max_global_size = 65536
+    global_size = min(((shots + max_local_size - 1) // max_local_size) * max_local_size, max_global_size)
+    local_size = max_local_size
+    num_groups = global_size // local_size
+
+    # Bit-packing params
+    words = (n + 31) // 32
+
+    # Random seeds (host)
+    rng_seeds_np = np.random.randint(1, 2**31-1, size=global_size, dtype=np.uint32)
+
+    # Device buffers
+    mf = cl.mem_flags
+    G_m_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=G_m_np)
+    thresholds_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=thresholds_np)
+    rng_buf = cl.Buffer(ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=rng_seeds_np)
+
+    # Solutions buffer: each work-item writes a candidate bitset
+    solutions_buf = cl.Buffer(ctx, mf.READ_WRITE, size=num_groups * words * np.uint32().nbytes)
+
+    # Best energies and solutions (per group)
+    # We'll reuse solutions_buf to hold them after reduction (kernel copies winners into group slices)
+    best_energies_np = np.empty(num_groups, dtype=np.float32)
+    best_energies_buf = cl.Buffer(ctx, mf.WRITE_ONLY, best_energies_np.nbytes)
+
+    # Local memory buffers
+    local_energy_buf = cl.LocalMemory(np.dtype(np.float32).itemsize * local_size)
+    local_index_buf = cl.LocalMemory(np.dtype(np.int32).itemsize * local_size)
+
+    # Set kernel args
+    kernel.set_args(
+        G_m_buf,
+        thresholds_buf,
+        np.int32(n),
+        np.int32(shots),
+        rng_buf,
+        solutions_buf,
+        best_energies_buf,
+        local_energy_buf,
+        local_index_buf
+    )
+
+    # Launch
+    cl.enqueue_nd_range_kernel(queue, kernel, (global_size,), (local_size,))
+
+    # Copy back
+    cl.enqueue_copy(queue, best_energies_np, best_energies_buf)
+    solutions_np = np.empty(num_groups * words, dtype=np.uint32)
+    cl.enqueue_copy(queue, solutions_np, solutions_buf)  # all candidates, includes per-group winners
+    queue.finish()
+
+    # Host reduction
+    best_group = np.argmax(best_energies_np)  # max cut
+    best_energy = float(best_energies_np[best_group])
+    best_solution_bits = solutions_np[best_group * words : (best_group + 1) * words]
+
+    # Unpack bitset into boolean vector
+    best_solution = np.zeros(n, dtype=np.bool_)
+    for u in range(n):
+        w = u >> 5
+        b = u & 31
+        best_solution[u] = (best_solution_bits[w] >> b) & 1
+
+    return best_solution, best_energy
+
+
 @njit
 def cpu_footer(shots, quality, n_qubits, G_m, nodes):
     J_eff, degrees = init_J_and_z(G_m)
@@ -145,12 +214,7 @@ def cpu_footer(shots, quality, n_qubits, G_m, nodes):
 
 @njit
 def gpu_footer(shots, n_qubits, G_m, J_eff, hamming_prob, nodes):
-    hamming_prob /= hamming_prob.sum()
-    tot_prob = 0.0
-    for i in range(n_qubits - 1):
-        tot_prob += hamming_prob[i]
-        hamming_prob[i] = tot_prob
-    hamming_prob[-1] = 2.0
+    fix_cdf(hamming_prob)
 
     best_solution, best_value = sample_for_solution(G_m, shots, hamming_prob, J_eff)
 
@@ -162,7 +226,7 @@ def gpu_footer(shots, n_qubits, G_m, J_eff, hamming_prob, nodes):
 def maxcut_tfim(
     G,
     quality=None,
-    shots=None,
+    shots=None
 ):
     nodes = None
     n_qubits = 0
@@ -177,33 +241,61 @@ def maxcut_tfim(
         G_m = G
 
     if n_qubits < 3:
-        return low_width(G_m, nodes, n_qubits)
+        if n_qubits == 0:
+            return "", 0, ([], [])
+
+        if n_qubits == 1:
+            return "0", 0, (nodes, [])
+
+        if n_qubits == 2:
+            weight = G_m[0, 1]
+            if weight < 0.0:
+                return "00", 0, (nodes, [])
+
+            return "01", weight, ([nodes[0]], [nodes[1]])
 
     if quality is None:
-        quality = 8
+        quality = 3
 
     if shots is None:
         # Number of measurement shots
         shots = n_qubits << quality
 
-    n_steps = 1 << quality
+    n_steps = 2 << quality
     grid_size = n_steps * n_qubits
 
     if not (IS_OPENCL_AVAILABLE and grid_size >= 128):
         return cpu_footer(shots, quality, n_qubits, G_m, nodes)
 
     J_eff, degrees = init_J_and_z(G_m)
-    hamming_prob = init_thresholds(n_qubits)
 
     delta_t = 1.0 / n_steps
     tot_t = 2.0 * n_steps * delta_t
     h_mult = 2.0 / tot_t
-    theta = init_theta(delta_t, tot_t, h_mult, n_qubits, J_eff, degrees)
+
     args = np.empty(3, dtype=np.float32)
     args[0] = delta_t
     args[1] = tot_t
     args[2] = h_mult
 
+    # Move to GPU
+    mf = cl.mem_flags
+    args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
+    J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
+    deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
+    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE, size=(n_qubits * 8))
+
+    # Warp size is 32:
+    group_size = min(n_qubits, 64)
+    global_size = ((n_qubits + group_size - 1) // group_size) * group_size
+
+    opencl_context.init_theta_kernel(
+        opencl_context.queue, (global_size,), (group_size,),
+        args_buf, np.int32(n_qubits), J_buf, deg_buf, theta_buf
+    )
+
+    hamming_prob = init_thresholds(n_qubits)
+
     # Warp size is 32:
     group_size = n_qubits - 1
     if group_size > 256:
@@ -211,11 +303,6 @@ def maxcut_tfim(
     grid_dim = n_steps * n_qubits * group_size
 
     # Move to GPU
-    mf = cl.mem_flags
-    args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
-    J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
-    deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
-    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta)
     ham_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=hamming_prob)
 
     # Kernel execution
@@ -226,5 +313,16 @@ def maxcut_tfim(
 
     # Fetch results
     cl.enqueue_copy(opencl_context.queue, hamming_prob, ham_buf)
+    opencl_context.queue.finish()
+
+    args_buf.release()
+    J_buf.release()
+    deg_buf.release()
+    theta_buf.release()
+
+    args_buf = None
+    J_buf = None
+    deg_buf = None
+    theta_buf = None
 
     return gpu_footer(shots, n_qubits, G_m, J_eff, hamming_prob, nodes)
