@@ -3,8 +3,9 @@ import networkx as nx
 import numpy as np
 import os
 from numba import njit, prange
+from scipy.sparse import lil_matrix, csr_matrix
 
-from .maxcut_tfim_util import get_cut, init_theta, init_thresholds, low_width, maxcut_hamming_cdf, opencl_context, probability_by_hamming_weight
+from .maxcut_tfim_util import fix_cdf, get_cut, init_thresholds, maxcut_hamming_cdf, opencl_context, probability_by_hamming_weight
 
 IS_OPENCL_AVAILABLE = True
 try:
@@ -44,56 +45,53 @@ def local_repulsion_choice(G_cols, G_data, G_rows, max_weight, weights, n, m):
     """
 
     weights = weights.copy()
-    chosen = np.zeros(m, dtype=np.int32)   # store chosen indices
-    available = np.ones(n, dtype=np.bool_) # True = available, False = not
-    mask = np.zeros(n, dtype=np.bool_)
-    chosen_count = 0
+    used = np.zeros(n, dtype=np.bool_) # False = available, True = used
 
     for _ in range(m):
         # Count available
         total_w = 0.0
         for i in range(n):
-            if available[i]:
-                total_w += weights[i]
-        if total_w <= 0:
-            break
+            if used[i]:
+                continue
+            total_w += weights[i]
 
         # Normalize & sample
         r = np.random.rand()
         cum = 0.0
         node = -1
         for i in range(n):
-            if available[i]:
-                cum += weights[i] / total_w
-                if r < cum:
-                    node = i
-                    break
+            if used[i]:
+                continue
+            cum += weights[i]
+            if (total_w * r) < cum:
+                node = i
+                break
 
         if node == -1:
-            continue
+            node = 0
+            while used[node]:
+                node += 1
 
         # Select node
-        chosen[chosen_count] = node
-        chosen_count += 1
-        available[node] = False
-        mask[node] = True
+        used[node] = True
 
         # Repulsion: penalize neighbors
         for j in range(G_rows[node], G_rows[node + 1]):
             nbr = G_cols[j]
-            if available[nbr]:
-                weights[nbr] *= 0.5 ** (G_data[j] / max_weight)  # tunable penalty factor
+            if used[nbr]:
+                continue
+            weights[nbr] *= max(2e-7, 1 - G_data[j] / max_weight)
 
         for nbr in range(node):
-            if not available[nbr]:
+            if used[nbr]:
                 continue
             start = G_rows[nbr]
             end = G_rows[nbr + 1]
             j = binary_search(G_cols[start:end], node) + start
             if j < end:
-                weights[nbr] *= 0.5 ** (G_data[j] / max_weight)  # tunable penalty factor
+                weights[nbr] *= max(2e-7, 1 - G_data[j] / max_weight)
 
-    return mask
+    return used
 
 
 @njit
@@ -112,7 +110,7 @@ def compute_energy(sample, G_data, G_rows, G_cols):
 def sample_for_solution(G_data, G_rows, G_cols, shots, thresholds, J_eff):
     n = G_rows.shape[0] - 1
     max_weight = G_data.max()
-    weights = 1.0 / (1.0 + (2 ** -52) - J_eff)
+    weights = (1.0 / (1.0 + (2e-52) - J_eff)).astype(np.float64)
 
     solutions = np.empty((shots, n), dtype=np.bool_)
     energies = np.empty(shots, dtype=np.float32)
@@ -191,12 +189,7 @@ def cpu_footer(shots, quality, n_qubits, G_data, G_rows, G_cols, nodes):
 
 @njit
 def gpu_footer(shots, n_qubits, G_data, G_rows, G_cols, J_eff, hamming_prob, nodes):
-    hamming_prob /= hamming_prob.sum()
-    tot_prob = 0.0
-    for i in range(n_qubits - 1):
-        tot_prob += hamming_prob[i]
-        hamming_prob[i] = tot_prob
-    hamming_prob[-1] = 2.0
+    fix_cdf(hamming_prob)
 
     best_solution, best_value = sample_for_solution(G_data, G_rows, G_cols, shots, hamming_prob, J_eff)
 
@@ -235,33 +228,61 @@ def maxcut_tfim_sparse(
         G_m = G
 
     if n_qubits < 3:
-        return low_width(G_m, nodes, n_qubits)
+        if n_qubits == 0:
+            return "", 0, ([], [])
+
+        if n_qubits == 1:
+            return "0", 0, (nodes, [])
+
+        if n_qubits == 2:
+            weight = G_m[0, 1]
+            if weight < 0.0:
+                return "00", 0, (nodes, [])
+
+            return "01", weight, ([nodes[0]], [nodes[1]])
 
     if quality is None:
-        quality = 8
+        quality = 3
 
     if shots is None:
         # Number of measurement shots
         shots = n_qubits << quality
 
-    n_steps = 1 << quality
+    n_steps = 2 << quality
     grid_size = n_steps * n_qubits
 
     if not (IS_OPENCL_AVAILABLE and grid_size >= 128):
         return cpu_footer(shots, quality, n_qubits, G_m.data, G_m.indptr, G_m.indices, nodes)
 
     J_eff, degrees = init_J_and_z(G_m.data, G_m.indptr, G_m.indices)
-    hamming_prob = init_thresholds(n_qubits)
 
     delta_t = 1.0 / n_steps
     tot_t = 2.0 * n_steps * delta_t
     h_mult = 2.0 / tot_t
-    theta = init_theta(delta_t, tot_t, h_mult, n_qubits, J_eff, degrees)
+
     args = np.empty(3, dtype=np.float32)
     args[0] = delta_t
     args[1] = tot_t
     args[2] = h_mult
 
+    # Move to GPU
+    mf = cl.mem_flags
+    args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
+    J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
+    deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
+    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE, size=(n_qubits * 4))
+
+    # Warp size is 32:
+    group_size = min(n_qubits, 64)
+    global_size = ((n_qubits + group_size - 1) // group_size) * group_size
+
+    opencl_context.init_theta_kernel(
+        opencl_context.queue, (global_size,), (group_size,),
+        args_buf, np.int32(n_qubits), J_buf, deg_buf, theta_buf
+    )
+
+    hamming_prob = init_thresholds(n_qubits)
+
     # Warp size is 32:
     group_size = n_qubits - 1
     if group_size > 256:
@@ -269,11 +290,6 @@ def maxcut_tfim_sparse(
     grid_dim = n_steps * n_qubits * group_size
 
     # Move to GPU
-    mf = cl.mem_flags
-    args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
-    J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
-    deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
-    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta)
     ham_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=hamming_prob)
 
     # Kernel execution
@@ -284,5 +300,16 @@ def maxcut_tfim_sparse(
 
     # Fetch results
     cl.enqueue_copy(opencl_context.queue, hamming_prob, ham_buf)
+    opencl_context.queue.finish()
+
+    args_buf.release()
+    J_buf.release()
+    deg_buf.release()
+    theta_buf.release()
+
+    args_buf = None
+    J_buf = None
+    deg_buf = None
+    theta_buf = None
 
     return gpu_footer(shots, n_qubits, G_m.data, G_m.indptr, G_m.indices, J_eff, hamming_prob, nodes)
