@@ -1,4 +1,5 @@
 from .maxcut_tfim_streaming import maxcut_tfim_streaming
+from .spin_glass_solver_util import get_cut_from_bit_array, int_to_bitstring
 import itertools
 import networkx as nx
 import numpy as np
@@ -24,8 +25,8 @@ def compute_energy(theta_bits, G_func, G_func_args_tuple, nodes):
     energy = 0
     for u in range(n_qubits):
         for v in range(u + 1, n_qubits):
-            eigen = 1 if theta_bits[u] == theta_bits[v] else -1
-            energy += G_func((nodes[u], nodes[v]), G_func_args_tuple) * eigen
+            val = G_func((nodes[u], nodes[v]), G_func_args_tuple)
+            energy += (val if theta_bits[u] == theta_bits[v] else -val)
 
     return energy
 
@@ -41,19 +42,22 @@ def bootstrap_worker(theta, G_func, G_func_args_tuple, nodes, indices):
 
 
 @njit(parallel=True)
-def bootstrap(theta, G_func, G_func_args_tuple, nodes, k, indices_array):
-    n = theta.shape[0]
+def bootstrap(best_theta, G_func, G_func_args_tuple, nodes, indices_array, k, min_energy):
+    n = len(indices_array) // k
     energies = np.empty(n, dtype=np.float32)
     for i in prange(n):
         j = i * k
-        energies[i] = bootstrap_worker(theta, G_func, G_func_args_tuple, nodes, indices_array[j : j + k])
+        energies[i] = bootstrap_worker(best_theta, G_func, G_func_args_tuple, nodes, indices_array[j : j + k])
 
-    return energies
+    energy = energies.min()
+    if energy < min_energy:
+        index_match = np.random.choice(np.where(energies == energy)[0])
+        indices = indices_array[(index_match * k) : ((index_match + 1) * k)]
+        min_energy = energy
+        for i in indices:
+            best_theta[i] = not best_theta[i]
 
-
-# By Gemini (Google Search AI)
-def int_to_bitstring(integer, length):
-    return (bin(integer)[2:].zfill(length))[::-1]
+    return min_energy
 
 
 def spin_glass_solver_streaming(
@@ -81,7 +85,7 @@ def spin_glass_solver_streaming(
             return "01", weight, ([nodes[0]], [nodes[1]]), -weight
 
     if quality is None:
-        quality = 6
+        quality = 3
 
     bitstring = ""
     if isinstance(best_guess, str):
@@ -97,6 +101,7 @@ def spin_glass_solver_streaming(
     min_energy = compute_energy(best_theta, G_func, G_func_args_tuple, nodes)
     improved = True
     correction_quality = 1
+    combos_list = []
     while improved:
         improved = False
         k = 1
@@ -104,20 +109,19 @@ def spin_glass_solver_streaming(
             if n_qubits < k:
                 break
 
-            theta = best_theta.copy()
+            combos = []
+            if len(combos_list) < k:
+                combos = np.array(list(
+                    item for sublist in itertools.combinations(range(n_qubits), k) for item in sublist
+                ))
+                combos_list.append(combos)
+            else:
+                combos = combos_list[k - 1]
 
-            combos = list(
-                item for sublist in itertools.combinations(range(n_qubits), k) for item in sublist
-            )
-            energies = bootstrap(theta, G_func, G_func_args_tuple, nodes, k, combos)
+            energy = bootstrap(best_theta, G_func, G_func_args_tuple, nodes, combos, k, min_energy)
 
-            energy = energies.min()
             if energy < min_energy:
-                index_match = np.random.choice(np.where(energies == energy)[0])
-                indices = combos[(index_match * k) : ((index_match + 1) * k)]
                 min_energy = energy
-                for i in indices:
-                    best_theta[i] = not best_theta[i]
                 improved = True
                 if correction_quality < (k + 1):
                     correction_quality = k + 1
@@ -125,17 +129,7 @@ def spin_glass_solver_streaming(
 
             k = k + 1
 
-    bitstring = ""
-    l, r = [], []
-    for i in range(len(best_theta)):
-        b = best_theta[i]
-        if b:
-            bitstring += "1"
-            r.append(nodes[i])
-        else:
-            bitstring += "0"
-            l.append(nodes[i])
-
+    bitstring, l, r = get_cut_from_bit_array(best_theta, nodes)
     cut_value = evaluate_cut_edges(best_theta, G_func, G_func_args_tuple, nodes)
 
     return bitstring, float(cut_value), (l, r), float(min_energy)
